unit freereverb;
{
  Unit: Reverb for either KOL or VCL Delphi and Freepascal 1.9.x,
  probably Kylix.
  purpose: Based on the Freeverb (C++) design by Jezar[at]dreampoint.co.uk
  Author: Thaddy de Koning, thaddy[at]thaddy.com
  Copyright: Original in C++ 2000, Jezar
  Delphi and Freepascal version 2003,
  Basm  added 2004, Thaddy de Koning
  Use as you like, copyrighted freeware.

  Remarks: Removed original skip for interleave.
  Comb and Allpass filter processing rewritten in BASM.
  It is as faithfull a translation as I could manage.
  Sound is exactly the same as FreeVerb 3.
  I have writen it to be as portable as my Pascals go
  Although not tested, it should work in BCB6+ and Kylix too.
  It works with Delphi 6/7 and Freepascal 1.9.X under both linux
  and windows.
  There are no dependencies on the windows unit.

  2014-05-16 : Changed to be used in WREN by Blue Hell to work on a single
  sample pair instead of a buffer. This was downloaded from :
  http://members.chello.nl/t.koning8/kolindex.htm .
  It will not work on all platforms anymore.
}

interface


uses

  Classes, KnobsUtils;


//
// Reverb model tuning values, taken from original algoritm by Jezar
//

const

  numcombs       = 8;
  numallpasses   = 4;
  muted          = 0;
  fixedgain      = 0.015;
  scalewet       = 3;
  scaledry       = 2;
  scaledamp      = 0.9;        // was 0.4 :: 0.9 provides more damping
  scaleroom      = 0.28;       // was 0.28
  offsetroom     = 0.7;
  initialroom    = 0.5;
  initialdamp    = 0.5;
  initialwet     = 1 / scalewet;
  initialdry     = 1 / scaledry;
  initialwidth   = 1;
  initialfreeze  = 0;
  freezemode     = 0.5;
  freezeamount   = 1.0;
  stereospread_l = 0;          // Original value
  stereospread_r = +23;        // Original value


type

  TAllpass = class( TObject)
  private
    feedback : TSignal;
    buffer   : TSignalArray;
    bufsize  : integer;
    bufidx   : integer;
  public
    constructor Create( Buffersize : integer);                                                                  virtual;
    function    process( const input : TSignal) : TSignal;
    procedure   mute;
    procedure   setfeedback( Value : TSignal);
    function    getfeedback: TSignal;
  end;


  Tcomb = class( TObject)
  private
    feedback    : TSignal;
    filterstore : TSignal;
    damp1       : TSignal;
    damp2       : TSignal;
    buffer      : TSignalArray;
    bufsize     : integer;
    bufidx      : integer;
  public
    constructor Create( Buffersize: Integer);                                                                   virtual;
    function    process( const input: TSignal): TSignal;
    procedure   mute;
    procedure   setdamp    ( Value : TSignal);
    function    getdamp    : TSignal;
    procedure   setfeedback( Value : TSignal);
    function    getfeedback: TSignal;
  end;


  TReverb = class( TObject)
  private
    gain      : TSignal;
    roomsize  : TSignal;
    roomsize1 : TSignal;
    damp      : TSignal;
    damp1     : TSignal;
    wet       : TSignal;
    wet1      : TSignal;
    wet2      : TSignal;
    dry       : TSignal;
    Width     : TSignal;
    mode      : TSignal;
    FFreeze   : TSignal;
    FSize     : TSignal;
    combL     : array[ 0 .. numcombs     - 1] of Tcomb;
    combR     : array[ 0 .. numcombs     - 1] of Tcomb;
    allpassL  : array[ 0 .. numallpasses - 1] of TAllpass;
    allpassR  : array[ 0 .. numallpasses - 1] of TAllpass;
  protected
    procedure   Update;
  public
    constructor Create( aSampleRate: Integer; aSize: TSignal);
    destructor  Destroy;                                                                                       override;
    procedure   Mute;
    procedure   Process( const anInL, anInR : TSignal; var anOutL, anOutR : TSignal);
    procedure   SetRoomSize( Value : TSignal);
    function    GetRoomSize: TSignal;
    procedure   SetDamp    ( Value : TSignal);
    function    GetDamp    : TSignal;
    procedure   SetWet     ( Value : TSignal);
    function    GetWet     : TSignal;
    procedure   SetDry     ( Value : TSignal);
    function    GetDry     : TSignal;
    procedure   SetWidth   ( Value : TSignal);
    function    GetWidth   : TSignal;
    procedure   SetMode    ( Value : TSignal);
    function    GetMode    : TSignal;
    procedure   SetFreeze  ( Value : TSignal);
    function    GetFreeze  : TSignal;
  end;


implementation


{ ========
  TAllpass = class( TObject)
  private
    feedback : TSignal;
    buffer   : pointer;
    bufsize  : integer;
    bufidx   : integer;
  public
}

    constructor TAllpass.Create( Buffersize: integer);
    begin
      inherited Create;
      bufsize := Buffersize * SizeOf( TSignal);
      SetLength( buffer, bufsize);
      bufidx  := 0;
    end;


    function TAllpass.getfeedback: TSignal;
    begin
      Result := feedback;
    end;


    procedure TAllpass.mute;
    begin
      Fillchar( buffer[ 0], bufsize * SizeOf( buffer[ 0]), 0);
    end;


    function TAllpass.process( const input: TSignal): TSignal;
    var
      BufOut : TSignal;
    begin
      BufOut := Normalize( buffer[ bufidx]);
      Result := - input + BufOut;
      buffer[ bufidx] := input + Normalize(( bufout * feedback));
      bufidx := ( bufidx + 1) mod bufsize;
    end;


    procedure TAllpass.setfeedback( Value : TSignal);
    begin
      feedback := Value;
    end;


{ ========
  Tcomb = class( TObject)
  private
    feedback    : TSignal;
    filterstore : TSignal;
    damp1       : TSignal;
    damp2       : TSignal;
    buffer      : pointer;
    bufsize     : integer;
    bufidx      : integer;
  public
}

    constructor Tcomb.Create( Buffersize : integer);
    begin
      inherited Create;
      bufsize     := Buffersize * SizeOf( TSignal);
      SetLength( buffer, bufsize);
      filterstore := 0;
      bufidx      := 0;
    end;


    function Tcomb.getdamp : TSignal;
    begin
      Result := damp1;
    end;

    function Tcomb.getfeedback : TSignal;
    begin
      Result := feedback;
    end;


    procedure Tcomb.mute;
    begin
      Fillchar( buffer[ 0], bufsize * SizeOf( buffer[ 0]), 0);
    end;


    function Tcomb.process( const input : TSignal) : TSignal;
    begin
      Result      := Normalize( buffer[ bufidx]);
      filterstore := Normalize( ( Result * damp2) + (filterstore * damp1));
      buffer[ bufidx] := input + Normalize( filterstore * feedback);
      bufidx := ( bufidx + 1) mod bufsize;
    end;


    procedure Tcomb.setdamp(Value : TSignal);
    begin
      damp1 := Value;
      damp2 := 1 - Value;
    end;


    procedure Tcomb.setfeedback(Value : TSignal);
    begin
      feedback := Value;
    end;


{ ========
  TReverb = class( TObject)
  private
    gain      : TSignal;
    roomsize  : TSignal;
    roomsize1 : TSignal;
    damp      : TSignal;
    damp1     : TSignal;
    wet       : TSignal;
    wet1      : TSignal;
    wet2      : TSignal;
    dry       : TSignal;
    Width     : TSignal;
    mode      : TSignal;
    FSize     : TSignal;
    combL     : array[ 0 .. numcombs     - 1] of Tcomb;
    combR     : array[ 0 .. numcombs     - 1] of Tcomb;
    allpassL  : array[ 0 .. numallpasses - 1] of TAllpass;
    allpassR  : array[ 0 .. numallpasses - 1] of TAllpass;
  protected
}

    procedure TReverb.Update;
    var
      i : integer;
    begin
      // Recalculate internal values after parameter change
      wet1 := wet * ( Width / 2 + 0.5);
      wet2 := wet * (( 1 - Width) / 2);

      if   mode >= freezemode
      then begin
        roomsize1 := Clip(        FFreeze, 0.0, 1.0);
        damp1     := Clip(  1.0 - FFreeze, 0.0, 1.0);
        gain      := Clip(( 1.0 - FFreeze) * fixedgain, 0.0, fixedgain);
      end
      else begin
        roomsize1 := roomsize;
        damp1     := damp;
        gain      := fixedgain;
      end;

      for i := 0 to numcombs - 1
      do begin
        combL[ i].setfeedback( roomsize1);
        combR[ i].setfeedback( roomsize1);
        combL[ i].setdamp( damp1);
        combR[ i].setdamp( damp1);
      end;
    end;


//  public

    constructor TReverb.Create( aSampleRate: Integer; aSize: TSignal);
    begin
      inherited Create;

      // These values assume 44.1KHz sample rate
      // they will probably be OK for 48KHz sample rate
      // but would need scaling for 96KHz (or other) sample rates.
      // The values were obtained by listening tests.
      // Blue Hell:
      // The values are now recalculated for different sample rates,
      // maybe some prime checking would be in order though

      if   aSize < 0.2
      then aSize := 0.2;

      FSize := aSize;

      combL   [ 0]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1116.0 / 44100.0) + stereospread_l)));
      combR   [ 0]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1116.0 / 44100.0) + stereospread_r)));
      combL   [ 1]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1188.0 / 44100.0) + stereospread_l)));
      combR   [ 1]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1188.0 / 44100.0) + stereospread_r)));
      combL   [ 2]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1277.0 / 44100.0) + stereospread_l)));
      combR   [ 2]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1277.0 / 44100.0) + stereospread_r)));
      combL   [ 3]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1356.0 / 44100.0) + stereospread_l)));
      combR   [ 3]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1356.0 / 44100.0) + stereospread_r)));
      combL   [ 4]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1422.0 / 44100.0) + stereospread_l)));
      combR   [ 4]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1422.0 / 44100.0) + stereospread_r)));
      combL   [ 5]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1491.0 / 44100.0) + stereospread_l)));
      combR   [ 5]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1491.0 / 44100.0) + stereospread_r)));
      combL   [ 6]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1557.0 / 44100.0) + stereospread_l)));
      combR   [ 6]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1557.0 / 44100.0) + stereospread_r)));
      combL   [ 7]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1617.0 / 44100.0) + stereospread_l)));
      combR   [ 7]    := Tcomb   .Create( NextPrime( Round(( aSize * aSampleRate * 1617.0 / 44100.0) + stereospread_r)));

      allpassL[ 0]    := TAllpass.Create( NextPrime( Round(( aSize * aSampleRate *  556.0 / 44100.0) + stereospread_l)));
      allpassR[ 0]    := TAllpass.Create( NextPrime( Round(( aSize * aSampleRate *  556.0 / 44100.0) + stereospread_r)));
      allpassL[ 1]    := TAllpass.Create( NextPrime( Round(( aSize * aSampleRate *  441.0 / 44100.0) + stereospread_l)));
      allpassR[ 1]    := TAllpass.Create( NextPrime( Round(( aSize * aSampleRate *  441.0 / 44100.0) + stereospread_r)));
      allpassL[ 2]    := TAllpass.Create( NextPrime( Round(( aSize * aSampleRate *  341.0 / 44100.0) + stereospread_l)));
      allpassR[ 2]    := TAllpass.Create( NextPrime( Round(( aSize * aSampleRate *  341.0 / 44100.0) + stereospread_r)));
      allpassL[ 3]    := TAllpass.Create( NextPrime( Round(( aSize * aSampleRate *  225.0 / 44100.0) + stereospread_l)));
      allpassR[ 3]    := TAllpass.Create( NextPrime( Round(( aSize * aSampleRate *  225.0 / 44100.0) + stereospread_r)));

      // Set default values

      allpassL[ 0].setfeedback( 0.5);
      allpassR[ 0].setfeedback( 0.5);
      allpassL[ 1].setfeedback( 0.5);
      allpassR[ 1].setfeedback( 0.5);
      allpassL[ 2].setfeedback( 0.5);
      allpassR[ 2].setfeedback( 0.5);
      allpassL[ 3].setfeedback( 0.5);
      allpassR[ 3].setfeedback( 0.5);

      SetWet     ( initialwet   );
      SetRoomSize( initialroom  );
      SetDry     ( initialdry   );
      SetDamp    ( initialdamp  );
      SetWidth   ( initialwidth );
      SetMode    ( initialfreeze);
      Mute;
    end;


    destructor  TReverb.Destroy;
    var
      i : integer;
    begin
      for i := Low( allPassL) to High( allPassL) do allpassL[ i].DisposeOf;
      for i := Low( allPassR) to High( allPassR) do allpassR[ i].DisposeOf;
      for i := Low( combL   ) to High( combL   ) do combL   [ i].DisposeOf;
      for i := Low( combR   ) to High( combR   ) do combR   [ i].DisposeOf;

      inherited;
    end;


    function    TReverb.GetDamp: TSignal;
    begin
      Result := damp / scaledamp;
    end;


    function    TReverb.GetDry: TSignal;
    begin
      Result := dry / scaledry;
    end;


    function    TReverb.GetMode: TSignal;
    begin
      if   mode >= freezemode
      then Result := 1
      else Result := 0;
    end;


    procedure   TReverb.SetFreeze( Value : TSignal);
    begin
      FFreeze := Value;
    end;


    function    TReverb.GetFreeze: TSignal;
    begin
      Result := FFreeze;
    end;


    function    TReverb.GetRoomSize : TSignal;
    begin
      Result := (roomsize - offsetroom) / scaleroom;
    end;


    function TReverb.GetWet : TSignal;
    begin
      Result := wet / scalewet;
    end;


    function TReverb.GetWidth : TSignal;
    begin
      Result := Width;
    end;


    procedure TReverb.Mute;
    var
      i : integer;
    begin
      if   GetMode >= freezemode
      then Exit;

      for i := 0 to numcombs - 1
      do begin
        combL[ i].mute;
        combR[ i].mute;
      end;
      for i := 0 to numallpasses - 1
      do begin
        allpassL[ i].mute;
        allpassR[ i].mute;
      end;
    end;


    procedure TReverb.Process(const anInL, anInR : TSignal; var anOutL, anOutR : TSignal);
    var
      outL, outR, input : TSignal;
      j                 : integer;
    begin
      outL  := 0;
      outR  := 0;
      input := ( anInL + anInR) * gain;

      // Accumulate comb filters in parallel
      for j := 0 to numcombs - 1
      do begin
        outL := outL + combL[ j].process( input);
        outR := outR + combR[ j].process( input);
      end;

      // Feed through allpasses in series
      for j := 0 to numallpasses - 1
      do begin
        outL := allpassL[ j].process( outL);
        outR := allpassR[ j].process( outR);
      end;

      // Calculate output
      anOutL := outL * wet1 + outR * wet2 + anInL * dry;
      anOutR := outR * wet1 + outL * wet2 + anInR * dry;
    end;


    procedure TReverb.SetDamp( Value : TSignal);
    var
      aNewValue: TSignal;
    begin
      aNewValue := Value * scaledamp;

      if   aNewValue <> damp
      then begin
        damp := aNewValue;
        Update;
      end;
    end;


    procedure TReverb.SetDry( Value : TSignal);
    var
      aNewValue: TSignal;
    begin
      aNewValue := Value * scaledry;

      if   aNewValue <> dry
      then begin
        dry := aNewValue;
        Update;
      end;
    end;


    procedure TReverb.SetMode( Value : TSignal);
    begin
      if   Value <> mode
      then begin
        mode := Value;
        Update;
      end;
    end;


    procedure TReverb.SetRoomSize( Value : TSignal);
    var
      aNewValue: TSignal;
    begin
      aNewValue := ( Value * scaleroom) + offsetroom;

      if   aNewValue <> roomsize
      then begin
        roomsize := aNewValue;
        Update;
      end;
    end;


    procedure TReverb.SetWet( Value : TSignal);
    var
      aNewValue: TSignal;
    begin
      aNewValue := Value * scalewet;

      if   aNewValue <> wet
      then begin
        wet := aNewValue;
        Update;
      end;
    end;


    procedure TReverb.SetWidth( Value : TSignal);
    begin
      if   Value <> width
      then begin
        Width := Value;
        Update;
      end;
    end;


end.

